import java.time.format.DateTimeFormatter
import java.time.LocalDate
import org.apache.commons.io.IOUtils

tempo = new TempoTool(logger: logger, apiToken: tempoApiToken, baseUrl: baseurl, dateFrom: tempoDateFrom, dateTo: tempoDateTo, issues: issues)

/**
 * Velocity tool to get Tempo worklogs from the Tempo Cloud Servlet API.
 *
 * @see https://tempo-io.github.io/tempo-mtm-docs/
 */
class TempoTool {
	def logger

	/* Tempo servlet request parameters. */
	def apiToken
	def baseUrl

	def dateFrom
	def dateTo

	def issues

	/**
	 * Whether to preload all the worklogs in the given timeframe with a single Tempo API call.
	 * Preloading performs better for exports with larger number of issues.
	 * If false, worklogs will be downloaded for each issue separately.
	 */
	def preloadWorklogs
	def preloadedWorklogs = []

	def worklogDateFormat = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss")

	def slurper = new XmlSlurper()

	/**
	 * Returns the Tempo managed worklogs for the passed issue as a <code>List</code> of <code>Maps</code>, or the exception object if something went wrong.
	 * Get the values by using the keys "hours", "work_date" and such, according to the documentation linked above.
	 */
	def getWorklogs(def issueKey) {
		preloadWorklogs = (issues.size() > 10)
		def result
		if (preloadWorklogs) {
			if (!preloadedWorklogs) {
				logger.log("Downloading Tempo worklogs in bulk")
				preloadedWorklogs = callTempoApi()
			}
			result = !(preloadedWorklogs instanceof Throwable) ? preloadedWorklogs.findAll { it.issue_key == issueKey } : preloadedWorklogs
		} else {
			logger.log("Downloading Tempo worklogs for ${issueKey}")
			result = callTempoApi(issueKey)
		}
		return !(result instanceof Throwable) ? sortWorklogsByDate(result) : result
	}

	/**
	 * Calls the Tempo API and returns a list of worklog objects (with their properties precisely mapping from the XML response).
	 * @param issueKey optional, if not defined, then all worklogs in the given timeframe are returned.
	 */
	def callTempoApi(def issueKey) {
		InputStream stream
		try {
			def url = "https://app.tempo.io/api/1/getWorklog?dateFrom=${dateFrom}&dateTo=${dateTo}&format=xml&addBillingInfo=true&addUserDetails=true&baseUrl=${baseUrl}&tempoApiToken=${apiToken}"
			if (issueKey) {
				url += "&issueKey=${issueKey}"
			}
			logger.log("Tempo URL" + (issueKey ? " for ${issueKey}" : "") + ": ${url}")
			stream = new URL(url).openStream()
			def xml = IOUtils.toString(stream, "UTF-8")
			logger.log("Tempo XML" + (issueKey ? " for ${issueKey}" : "") + ":")
			xml.eachLine { logger.log(it) }

			def result = slurper.parseText(xml)
			def worklogs = result.worklog
			logger.log("${worklogs.size()} Tempo worklogs found" + (issueKey ? " for ${issueKey}" : ""))

			def worklogObjects = []
			result.worklog.each { w ->
				worklogObjects.add(w)
			}
			logger.log("Worklogs" + (issueKey ? " for ${issueKey}" : "") + ": ${worklogObjects}")
			return worklogObjects
		} catch (Throwable e) {
			logger.log("Failed to download Tempo worklogs", e)
			return e
		} finally {
			IOUtils.closeQuietly(stream)
		}
	}

	/**
	 * Returns the passed worklogs sorted by date ascending.
	 * (Earlier dated worklogs will come first. For identical dates, the worklog with the less ID will come first.)
	 */
	def sortWorklogsByDate(worklogs) {
		return worklogs.sort { a, b ->
			def dateA = LocalDate.parse(a.work_date_time.text(), worklogDateFormat)
			def dateB = LocalDate.parse(b.work_date_time.text(), worklogDateFormat)
			if (dateA == dateB) {
				return a.worklog_id.text() as Long <=> b.worklog_id.text() as Long
			}
			return dateA <=> dateB
		}
	}
}
